/**
 * EditCardGUI.java
 * Group 7, Object Oriented Software Engineering
 * October 10, 2009
 */
package edu.jhu.cs.oose.elevel.AdvancedEditCard;

import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import com.trolltech.qt.QSignalEmitter;
import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QAction;
import com.trolltech.qt.gui.QCloseEvent;
import com.trolltech.qt.gui.QColor;
import com.trolltech.qt.gui.QColorDialog;
import com.trolltech.qt.gui.QFontDialog;
import com.trolltech.qt.gui.QGridLayout;
import com.trolltech.qt.gui.QGroupBox;
import com.trolltech.qt.gui.QInputDialog;
import com.trolltech.qt.gui.QLabel;
import com.trolltech.qt.gui.QMainWindow;
import com.trolltech.qt.gui.QMenu;
import com.trolltech.qt.gui.QMenuBar;
import com.trolltech.qt.gui.QMessageBox;
import com.trolltech.qt.gui.QVBoxLayout;
import com.trolltech.qt.gui.QWidget;
import edu.jhu.cs.oose.elevel.MainWindow;
import edu.jhu.cs.oose.elevel.dataModel.Card;
import edu.jhu.cs.oose.elevel.editing.EditCardWindow;
import edu.jhu.cs.oose.elevel.utilities.PreferenceCheckBox;

/**
 * GUI responsible for the view and controller for the underlying {@link EditCardWindow} model
 * 
 * @author Jesse Yates
 */
public class EditCardGUI extends QMainWindow
{
	/** The underlying {@link EditCardWindow} model */
	private EditCardWindow editWindow;

	/** The <code>Card</code> to be edited */
	@SuppressWarnings("unused")
	private Card card;

	/** The Question in displayable format */
	private DrawingGraphicsView question;
	private DrawingGraphicsScene questionModel;

	/** The answer in displayable format */
	private DrawingGraphicsView answer;
	private DrawingGraphicsScene answerModel;

	/** The hint in displayable format */
	private FocusingTextEdit hint;

	/** The hint in displayable format */
	private FocusingTextEdit addInfo;

	/** A {@link Map} representing the preferences of the <code>Card</code> */
	private Map<String, Boolean> pref;

	/**
	 * Physical widget holding the buttons
	 */
	private QGroupBox drawingButtons;

	/**
	 * Virtual construct grouping button actions together
	 */
	private DrawingButtonGroup toolButtons;

	/** A QMenu representing the file menu */
	private QMenu fileMenu;

	/** A QMenu representing the edit menu */
	private QMenu editMenu;

	/** A QMenu representing the help menu */
	private QMenu helpMenu;

	/** A QAction that displays the about dialog */
	private QAction aboutAct;

	/** A QAction that saves the current <code>Card</code> */
	private QAction saveAct;

	/** A QAction that closes the current dialog */
	private QAction closeAct;

	private QAction setFontAct;
	private QAction setPenWidthAct;
	private QAction setPenColorAct;

	private QAction undoAct;
	private QAction redoAct;

	/** Signal that the window is closing */
	public final QSignalEmitter.Signal0 closing = new QSignalEmitter.Signal0();

	/**
	 * Constructor for the GUI
	 * 
	 * @param parent
	 *        a {@link MainWindow} representing the main window of the application
	 * @param editWindow
	 *        a {@link EditCardWindow} representing the underlying model for the GUI
	 */
	public EditCardGUI(MainWindow parent, EditCardWindow editWindow)
	{
		super(parent);
		this.setMenuBar(new QMenuBar());
		this.editWindow = editWindow;
		this.pref = new HashMap<String, Boolean>();

		// setup the window for display
		this.setUpCard();
		this.createToolbar();

		QWidget editingArea = new QWidget();
		QGridLayout layout = new QGridLayout(editingArea);
		editingArea.setLayout(layout);

		layout.setHorizontalSpacing(3);
		layout.setVerticalSpacing(3);
		layout.setMargin(5);
		QVBoxLayout leftSide = new QVBoxLayout();
		QVBoxLayout rightSide = new QVBoxLayout();
		layout.addLayout(leftSide, 0, 0);
		layout.addLayout(rightSide, 0, 1);

		QVBoxLayout settingsLayout = new QVBoxLayout();
		settingsLayout.addWidget(new QLabel("Settings"));
		for(Entry<String, Boolean> i : this.pref.entrySet())
		{
			PreferenceCheckBox box = new PreferenceCheckBox(i.getKey(), i.getKey());
			if(i.getValue().booleanValue())
				box.setCheckState(Qt.CheckState.Checked);
			box.stateChange.connect(this, "updateSetting(String, boolean)");
			settingsLayout.addWidget(box);
		}

		layout.setColumnMinimumWidth(0, 100);
		layout.setColumnMinimumWidth(1, 50);
		layout.setRowMinimumHeight(0, 200);

		leftSide.addWidget(new QLabel("Question"), 0, Qt.AlignmentFlag.AlignCenter);
		leftSide.addWidget(this.question);

		leftSide.addWidget(new QLabel("Answer"), 0, Qt.AlignmentFlag.AlignCenter);
		leftSide.addWidget(this.answer);

		rightSide.setSpacing(10);
		rightSide.addWidget(new QLabel("Hint"), 0, Qt.AlignmentFlag.AlignCenter);
		rightSide.addWidget(this.hint);
		rightSide.addWidget(new QLabel("Additional Information"), 0, Qt.AlignmentFlag.AlignCenter);
		rightSide.addWidget(this.addInfo);
		rightSide.addLayout(settingsLayout);
		rightSide.setAlignment(settingsLayout, Qt.AlignmentFlag.AlignCenter);

		this.setCentralWidget(editingArea);

		setupActions();
		setupMenu();
	}

	/**
	 * Display information about editing a <code>Card</code>
	 */
	public void about()
	{
		QMessageBox.information(this, tr("Editing a Card"), tr("Different toolbars allow for different editing. \nSimply put, input the question, answer, hint and additional information as you see fit, and then select the options for the card. Finally save the card."));
	}

	/**
	 * If the window attempts to close, determine if the card should be saved
	 * 
	 * @see com.trolltech.qt.gui.QWidget#closeEvent(com.trolltech.qt.gui.QCloseEvent)
	 */
	public void closeEvent(QCloseEvent event)
	{
		if(maybeSave())
		{
			event.accept();
			closing.emit();
		}
		else
			event.ignore();
	}

	/**
	 * Prompt the user to determine if the <code>Card</code> should be saved
	 * 
	 * @return <code>true</code> if the <code>Card</code> should be saved
	 */
	private boolean displaySavePrompt()
	{
		QMessageBox.StandardButton ret = QMessageBox.warning(this, tr("Edit Card"), "The card has been modified.\n" + "Save your changes?", new QMessageBox.StandardButtons(QMessageBox.StandardButton.Ok, QMessageBox.StandardButton.No, QMessageBox.StandardButton.Cancel));
		if(ret == QMessageBox.StandardButton.Ok)
			saveCard();
		else if(ret == QMessageBox.StandardButton.Cancel)
			return false;

		return true;
	}

	/**
	 * Returns the {@link EditCardWindow} model this GUI is associated with
	 * 
	 * @return the {@link EditCardWindow} model this GUI is associated with
	 */
	public EditCardWindow getEditWindow()
	{
		return this.editWindow;
	}

	/**
	 * Check to see the <code>card</code> should be saved
	 * 
	 * @return True if the <code>card</code> should be saved, false otherwise
	 */
	private boolean maybeSave()
	{
		Card original = this.editWindow.getCard();
		if(original.getAnswer().equals(this.answer.toString()))
			if(original.getQuestion().equals(this.question.toString()))
				if(original.getHint().equals(this.hint.toPlainText()))
					if(original.getAdditionalInfo().equals(this.addInfo.toPlainText()))
						if(this.pref.equals(original.getOptions()))
							return true;
		return displaySavePrompt();
	}

	/**
	 * Save all the current values to the <code>Card</code> Currently has an ans object for storing the text from the
	 * box or the image -> may or may not be used depending on time it takes to rebuild the scene and then display each
	 * item
	 */
	private void saveCard()
	{
		Object ans;
		if(this.pref.get("Self-check answers").equals(Boolean.TRUE))
			ans = this.answer.getImage();
		else
			ans = this.answer.getStrings();
		this.editWindow.save(this.question.toString(), this.answer.toString(), this.hint.toPlainText(), this.addInfo.toPlainText(), this.pref);
	}

	/**
	 * Initialize the actions in the menus
	 */
	private void setupActions()
	{
		this.saveAct = new QAction(tr("&Save"), this);
		this.saveAct.setShortcut(tr("Ctrl+S"));
		this.saveAct.setStatusTip(tr("Save the current card"));
		this.saveAct.triggered.connect(this, "saveCard()");

		this.closeAct = new QAction(tr("&Exit"), this);
		this.closeAct.setShortcut(tr("Ctrl+W"));
		this.closeAct.triggered.connect(this, "close()");

		this.aboutAct = new QAction(tr("&About"), this);
		this.aboutAct.setShortcut(tr("Ctrl+A"));
		this.aboutAct.triggered.connect(this, "about()");

		this.undoAct = new QAction(tr("&Undo"), this);
		this.undoAct.setShortcut(tr("Ctrl+Z"));
		this.undoAct.triggered.connect(this.answerModel, "undo()");
		this.undoAct.triggered.connect(this.questionModel, "undo()");
		this.undoAct.triggered.connect(this.hint, "focusUndo()");
		this.undoAct.triggered.connect(this.addInfo, "focusUndo()");

		this.redoAct = new QAction(tr("&Redo"), this);
		this.redoAct.setShortcut(tr("Ctrl+R"));
		this.redoAct.triggered.connect(this.answerModel, "redo()");
		this.redoAct.triggered.connect(this.questionModel, "redo()");
		this.redoAct.triggered.connect(this.hint, "focusRedo()");
		this.redoAct.triggered.connect(this.addInfo, "focusRedo()");

		this.setPenColorAct = new QAction(tr("Pen Color"), this);
		this.setPenColorAct.triggered.connect(this, "setPenColor()");

		this.setPenWidthAct = new QAction(tr("Pen Width"), this);
		this.setPenWidthAct.triggered.connect(this, "setPenWidth()");

		this.setFontAct = new QAction(tr("Font"), this);
		this.setFontAct.triggered.connect(this, "setFont()");

	}

	/**
	 * Get the information from the current {@link Card}
	 */
	private void setUpCard()
	{
		this.hint = new FocusingTextEdit(this.editWindow.getCard().getHint());
		this.hint.setMinimumSize(20, 20);

		this.addInfo = new FocusingTextEdit(this.editWindow.getCard().getAdditionalInfo());
		this.addInfo.setMinimumSize(20, 20);

		this.answerModel = new DrawingGraphicsScene();
		this.answer = new DrawingGraphicsView(this.answerModel);// DrawingGraphicsView.createWithModel();//
		// this.editWindow.getCard().getAnswer();
		// this.answer.setSceneRect(new QRectF(0, 0, 100, 200));

		this.questionModel = new DrawingGraphicsScene();
		this.question = new DrawingGraphicsView(this.questionModel);// this.editWindow.getCard().getQuestion();
		// this.question.setSceneRect(new QRectF(0, 0, 100, 200));

		this.pref = new HashMap<String, Boolean>();
		for(Entry<String, Boolean> j : this.editWindow.getCard().getOptions().entrySet())
			this.pref.put(new String(j.getKey()), new Boolean(j.getValue()));

		this.answerModel.gainedFocus.connect(this, "enableDrawingPreferences()");
		this.questionModel.gainedFocus.connect(this, "enableDrawingPreferences()");

		this.hint.gainedFocus.connect(this, "disableDrawingPreferences()");
		this.hint.gainedFocus.connect(this, "disableDrawingPreferences()");
	}

	/**
	 * Initialize the menus
	 */
	private void setupMenu()
	{
		this.fileMenu = menuBar().addMenu(tr("&File"));
		this.fileMenu.addAction(this.saveAct);
		this.fileMenu.addSeparator();
		this.fileMenu.addAction(this.closeAct);

		this.editMenu = menuBar().addMenu(tr("&Edit"));
		this.editMenu.addAction(undoAct);
		this.editMenu.addAction(redoAct);
		this.editMenu.addSeparator();
		this.editMenu.addAction(this.setFontAct);
		this.editMenu.addAction(this.setPenColorAct);
		this.editMenu.addAction(this.setPenWidthAct);
		/*
		 * editMenu.addAction(cutAct);
		 * editMenu.addAction(copyAct);
		 * editMenu.addAction(pasteAct);
		 */
		menuBar().addSeparator();

		this.helpMenu = menuBar().addMenu(tr("&Help"));
		this.helpMenu.addAction(this.aboutAct);
	}

	private void createToolbar()
	{

		System.out.println("Creating toolbar");
		this.toolButtons = new DrawingButtonGroup();
		DefaultToolBar toolbar = new DefaultToolBar();
		this.drawingButtons = new QGroupBox("Drawing style");
		DrawingRadioButton text = new DrawingRadioButton("Text", DrawingType.TEXT);
		text.setChecked(true);
		DrawingRadioButton line = new DrawingRadioButton("Line", DrawingType.LINE);
		DrawingRadioButton circle = new DrawingRadioButton("Circle", DrawingType.CIRCLE);
		DrawingRadioButton square = new DrawingRadioButton("Square", DrawingType.SQUARE);
		QVBoxLayout layout = new QVBoxLayout();

		layout.addWidget(text);
		layout.addWidget(line);
		layout.addWidget(circle);
		layout.addWidget(square);

		this.toolButtons.addButton(text);
		this.toolButtons.addButton(line);
		this.toolButtons.addButton(circle);
		this.toolButtons.addButton(square);

		toolButtons.typeChange.connect(this.answer.scene(), "setType(DrawingType)");
		toolButtons.typeChange.connect(this.question.scene(), "setType(DrawingType)");

		// Set the default type to text
		toolButtons.typeChange.emit(DrawingType.TEXT);
		this.drawingButtons.setLayout(layout);
		toolbar.addWidget(this.drawingButtons);
		this.addToolBar(Qt.ToolBarArea.LeftToolBarArea, toolbar);

	}

	/**
	 * Update the stored setting
	 * 
	 * @param k
	 *        The setting name
	 * @param v
	 *        The value to store
	 */
	public void updateSetting(String k, boolean v)
	{
		this.pref.put(k, new Boolean(v));
	}

	private void setPenColor()
	{

		DrawingGraphicsScene model;
		if(this.answerModel.hasFocus() || this.questionModel.hasFocus())
		{
			System.out.println("Setting pen color!");
			if(this.answerModel.hasFocus())
			{
				model = this.answerModel;
				System.out.println("Model: answer");
			}
			else
			{
				model = this.questionModel;
				System.out.println("Model: question");
			}

			QColor newColor = QColorDialog.getColor(model.penColor());
			if(newColor.isValid())
				model.setPenColor(newColor);
		}
	}

	private void setPenWidth()
	{
		System.out.println("Setting pen width!");
		DrawingGraphicsScene model;

		if(this.answerModel.hasFocus() || this.questionModel.hasFocus())
		{
			System.out.println("Setting pen width!");
			if(this.answerModel.hasFocus())
			{
				System.out.println("Model: answer");
				model = this.answerModel;
			}
			else
			{
				System.out.println("Model: question");
				model = this.questionModel;
			}
			Integer newWidth = QInputDialog.getInteger(this, tr("Edit Card"), tr("Select pen width:"), model.penWidth(), 1, 50, 1);
			if(newWidth != null)
				model.setPenWidth(newWidth.intValue());
		}
	}

	private void setFont()
	{
		QWidget widget = this.focusWidget();
		System.out.println("Setting the font! from: " + widget.font());
		widget.setFont(QFontDialog.getFont(widget.font()).font);
		System.out.println("Setting font to: " + widget.font());
	}

	private void disableDrawingPreferences()
	{
		// this.setFontAct.setEnabled(false);
		this.setPenColorAct.setEnabled(false);
		this.setPenWidthAct.setEnabled(false);
	}

	private void enableDrawingPreferences()
	{
		// this.setFontAct.setEnabled(true);
		this.setPenColorAct.setEnabled(true);
		this.setPenWidthAct.setEnabled(true);
	}
}
