/**
 * DeckTreeGUI.java
 * Group 7, Object Oriented Software Engineering
 * Oct 25, 2009
 */
package edu.jhu.cs.oose.elevel;

import com.trolltech.qt.core.QModelIndex;
import com.trolltech.qt.gui.QTreeView;
import edu.jhu.cs.oose.elevel.dataModel.Card;
import edu.jhu.cs.oose.elevel.dataModel.Deck;
import edu.jhu.cs.oose.elevel.utilities.ViewState;

/**
 * It's actually a deck tree view, but it manages its model
 * so that the main window doesn't have to
 * 
 * @author Pablo Lee
 */
public class DeckTreeGUI extends QTreeView
{
	/** The tree model it works with to display the tree */
	private DeckTreeModel treeModel;

	/** A reference to the singleton ViewState */
	private ViewState state;

	/**
	 * Initializes the class by enabling the user to click or drag and drop objects
	 * in the tree GUI such as {@link Deck}s and {@link Card}s
	 */
	public DeckTreeGUI(DeckTreeModel treeModel)
	{
		this.treeModel = treeModel;
		this.state = ViewState.INSTANCE;

		this.state.cardAdded.connect(this, "expandAndSelect(Card, Integer)");
		this.state.deckAdded.connect(this, "selectDeck(Deck, Integer)");
		this.state.cardChangedInCardArea.connect(this, "selectAndRefreshCard(Card, Deck)");

		setHeaderHidden(true);

		setSelectionMode(SelectionMode.SingleSelection);
		setDragDropMode(DragDropMode.DragDrop);
		setDragEnabled(true);
		setAcceptDrops(true);
		setDropIndicatorShown(true);
	}

	/**
	 * Expands the deck the card belongs to, and 
	 * selects that card.
	 * 
	 * @param c The card to select
	 * @param i Not used
	 */
	@SuppressWarnings("unused")
	private void expandAndSelect(Card c, Integer i)
	{
		QModelIndex deckIndex = treeModel.indexOf(state.getCurrentDeck());
		expand(deckIndex);
		
		selectAndRefreshCard(c, state.getCurrentDeck());
	}

	/**
	 * Selects the deck
	 * @param d The deck that's selected
	 * @param index Not used
	 */
	private void selectDeck(Deck d, Integer index)
	{
		setCurrentIndex(treeModel.valueToIndex(d));
	}
	
	/**
	 * Selects and refreshes the card passed on.
	 * If the card is null, the deck is selected instead.
	 * 
	 * @param c The card to select and refresh
	 * @param d The deck to select if the card is not set
	 */
	private void selectAndRefreshCard(Card c, Deck d)
	{
		if(c == null)
		{
			selectDeck(d, 0);
			return;
		}
		
		QModelIndex cardIndex = treeModel.indexOf(c);
		setCurrentIndex(cardIndex);
		
		dataChanged(cardIndex, cardIndex);
	}
	
	/**
	 * Called whenever a new selection's been made in the model,
	 * whether through the GUI or through the model itself.
	 * a <code>deck</code> or a new <code>card</code> and updates all signals to make that object the current object
	 */
	protected void currentChanged(QModelIndex current, QModelIndex previous)
	{
		Object selectedItem = treeModel.indexToValue(current);

		Card card;
		Deck deck;

		if(treeModel.isCard(selectedItem))
		{
			card = (Card) selectedItem;
			deck = (Deck) treeModel.indexToValue(treeModel.parent(current));
			state.setCurrentDeck(deck);
			state.setCurrentCard(card, deck);
		}
		else
		{
			deck = (Deck) selectedItem;
			if(deck == null)
			{
				return;
			}
			
			state.setCurrentDeck(deck);
		}
	}
}
