/**
 * Card.java
 * Group 7, Object Oriented Software Engineering
 * October 10, 2009
 */
package edu.jhu.cs.oose.elevel.dataModel;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

/**
 * <code>Card</code> represents the high-level intuition of a flash card. <code>Card</code> represents a single card in
 * a {@link Deck} and maintains all the features of a real flash-card as well as additional, computer-only
 * possibilities.
 * 
 * @author Jesse Yates, H. Parker Shelton
 */
public class Card implements Serializable
{
	private static final long serialVersionUID = 67658268L;

	/** A <code>String</code> representing the question associated with this <code>Card</code> */
	public String question;

	/** A <code>String</code> representing the answer associated with this <code>Card</code> */
	public String answer;
	
	/** a <code>Integer</code> representing a unique identifier for this card. Note: this field is "Package protected" (when no explicit modifier is given). This is the desired behaviour for this field. We only want Card, Deck, and TestStat to be able to modify id.**/
	int id;

	/** A <code>String</code> representing a hint associated with this <code>Card</code> */
	private String hint;

	/** A <code>String</code> representing additional information associated with this </code>Card</code> */
	private String additionalInfo;

	/** A {@link Map} representing the card's options */
	private Map<String, Boolean> options;

	/**
	 * Instantiates a new Card
	 */
	public Card()
	{
		this.options = new HashMap<String, Boolean>();
		this.answer = new String();
		this.question = new String();
		this.hint = new String();
		this.additionalInfo = new String();
		setDefaultSettings();
	}

	/**
	 * Instantiates a <code>Card</code> with the specified parameters.
	 * 
	 * @param q
	 *        a <code>String</code> representing the question associated with this <code>Card</code>
	 * @param a
	 *        a <code>String</code> representing the answer associated with this <code>Card</code>
	 * @param h
	 *        a <code>String</code> representing the hint associated with this <code>Card</code>
	 * @param i
	 *        a <code>String</code> representing the additional information associated with this <code>Card</code>
	 */
	public Card( String q, String a, String h, String i)
	{
		this.question = q;
		this.answer = a;
		setHint(h);
		setAdditionalInfo(i);

		this.options = new HashMap<String, Boolean>();
		setDefaultSettings();
	}

	/**
	 * Compares the provided <code>String</code> with this <code>Card</code>'s correct answer
	 * 
	 * @param a
	 *        a <code>String</code> representing the provided answer
	 * @return <code>true</code> if the answer sent matches, <code>false</code> otherwise
	 */
	public boolean checkAnswer(String a)
	{
		return a.equalsIgnoreCase(this.answer);
	}

	/**
	 * a comparison on hash codes
	 * @param other the card to be compared to this card
	 * @return true if this card's hashcode equals other's hashcode, false otherwise.
	 */
	public boolean equals(Card other)
	{
		return hashCode() == other.hashCode();
	}
	
	/**
	 * gets the card's deck-dependent ID.
	 * @return the ID
	 */
	public int getID()
	{
		return id;
	}

	/**
	 * Returns the additional information associated with this card
	 * 
	 * @return a <code>String</code> representing the additional information associated with this <code>Card</code>
	 */
	public String getAdditionalInfo()
	{
		return this.additionalInfo;
	}

	/**
	 * Returns the answer associated with this <code>Card</code>
	 * 
	 * @return a <code>String</code> representing this <code>Card</code>'s answer
	 */
	public String getAnswer()
	{
		return this.answer;
	}

	/**
	 * Get this <code>Card</code>'s current hint
	 * 
	 * @return The current hint
	 */
	public String getHint()
	{
		return this.hint;
	}

	/**
	 * Gets the cards options
	 * @return A map from strings (options) to boolean values (enabled vs disabled).
	 */
	public Map<String, Boolean> getOptions()
	{
		return options;
	}

	/**
	 * Returns this <code>Card</code>'s question
	 * 
	 * @return the question
	 */
	public String getQuestion()
	{
		return this.question.toString();
	}
	
	/**
	 * Computes the hash code of this card based on its question, answer, hint, additional information, and options fields.
	 */
	public int hashCode()
	{
		return ((this.question != null) ? 97 * this.question.hashCode() : 0) + ((this.answer != null) ? 79 * this.answer.hashCode() : 0) + ((this.hint != null) ? 61 * this.hint.hashCode() : 0) + ((this.additionalInfo != null) ? 47 * this.additionalInfo.hashCode() : 0) + ((this.options != null) ? 3 * this.options.hashCode() : 0) + id;
	}

	/**
	 * Sets the additional information associated with this card to the specified <code>String</code>
	 * 
	 * @param i
	 *        a <code>String<code> representing the card's additional information
	 */
	public void setAdditionalInfo(String i)
	{
		if(i == null)
			this.additionalInfo = "";
		else
			this.additionalInfo = i;
	}

	/**
	 * Set the default <code>Card</code> settings
	 */
	private void setDefaultSettings()
	{
		this.options.clear();
		this.options.put("Display Hints", new Boolean(false));
		this.options.put("Self-check Answers", new Boolean(false));
	}

	/**
	 * Sets the hint associated with this card to the specified <code>String</code>
	 * 
	 * @param h
	 *        a <code>String</code> representing the card's hint
	 */
	public void setHint(String h)
	{
		if(h == null)
			this.hint = "";
		else
			this.hint = h;
	}

	/**
	 * Set the provided setting. Update the setting if it already exists.
	 * 
	 * @param key
	 *        The setting name
	 * @param value
	 *        the Boolean value
	 */
	public void setSetting(String key, Boolean value)
	{
		if(this.options != null)
			this.options.put(key, value);
	}

	/**
	 * Set the values of this <code>Card</code>
	 * 
	 * @param q
	 *        a <code>String</code> representing the question associated with this <code>Card</code>
	 * @param a
	 *        a <code>String</code> representing the answer associated with this <code>Card</code>
	 * @param h
	 *        a <code>String</code> representing the hint associated with this <code>Card</code>
	 * @param i
	 *        a <code>String</code> representing the additional information associated with this <code>Card</code>
	 * @param pref
	 *        A {@link Map} representing the options associated with this <code>Card</code>
	 */
	public void setValues(String q, String a, String h, String i, Map<String, Boolean> pref)
	{
		this.question = q;
		this.answer = a;
		setHint(h);
		setAdditionalInfo(i);

		for(Entry<String, Boolean> j : pref.entrySet())
			this.options.put(new String(j.getKey()), new Boolean(j.getValue()));
	}

	/**
	 * Writes out the properties of the card
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString()
	{
		StringBuffer b = new StringBuffer();
		b.append("-----\n");
		b.append("Card:\n");
		b.append("Question: " + this.question + "\n");
		b.append("Answer: " + this.answer + "\n");
		b.append("Additional Info: " + this.additionalInfo + "\n");
		b.append("Hint: " + this.hint + "\n");
		b.append("Settings: \n");

		for(Entry<String, Boolean> i : this.options.entrySet())
			b.append(i.getKey() + ":" + i.getValue().toString() + "\n");
		b.append("-----\n");
		return b.toString();
	}
}
