/**
 * Deck.java
 * Group 7, Object Oriented Software Engineering
 * October 10, 2009
 */
package edu.jhu.cs.oose.elevel.dataModel;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import com.trolltech.qt.gui.QMessageBox;
import com.trolltech.qt.gui.QWidget;

/**
 * {@link Deck} represents the high-level intuition of a deck of flash cards.
 * The class stores a set of {@link Card}s and set of statistics based on past tests.
 * 
 * @author Jesse Yates, Greg Anderson, H. Parker Shelton
 */
public class Deck implements Serializable
{
	private static final long serialVersionUID = 68696775L;

	/** The location of this <code>Deck</code> on disk */
	private String diskLocation;

	/** The name of this <code>Deck</code> */
	private String name;

	/** The {@link Card} current selected */
	private int currentCard;

	/** The set of statistics for this <code>Deck</code>, compiled over time */
	private List<TestStat> testStats;

	/** The set of cards contained in this <code>Deck</code> */
	private List<Card> cards;

	/** The value used for the id of the next card to be added to this deck. Incremented automatically. */
	private int cardIdIterator = 0;

	/** A boolean indicating whether the given Deck has been changed */
	private boolean hasChanged;

	/**
	 * Initialize a basic <code>Deck</code> using the name provided
	 * 
	 * @param name
	 *        A String representing the name of the <code>Deck</code>
	 */
	public Deck(String name)
	{
		setName(name);
		this.hasChanged = false;
		this.cards = new ArrayList<Card>();
		this.testStats = new ArrayList<TestStat>();
	}

	public Deck()
	{
		this("");
	}

	/**
	 * Add the specified {@link Card} to the <code>Deck</code>
	 * 
	 * @param c
	 *        a {@link Card} to be added to the deck
	 */
	public boolean addCard(Card c)
	{
		hasChanged = true;
		c.id = getUnusedID();
		return this.cards.add(c);
	}

	/**
	 * Add the list of cards to this deck
	 * 
	 * @param cards
	 *        A <code>List</code> of cards to add to this deck
	 * @return <code>true</code> if all cards were successfully added to the deck, <code>false</code> otherwise
	 */
	public boolean addCards(List<Card> cards)
	{
		hasChanged = true;
		this.cards.addAll(cards);
		return true;
	}

	/**
	 * Add the specified {@link Card} to the <code>Deck</code> at position <code>i</code>
	 * 
	 * @param i
	 *        An <code>int</code> representing the position at which to add the specified card
	 * @param c
	 *        The <code>Card</code> to be added
	 */
	public void addCardAtIndex(int i, Card c)
	{
		hasChanged = true;
		this.cards.add(i, c);
	}

	/**
	 * Adds a new testing statistic to the current set of statistics for the deck
	 * 
	 * @param d
	 *        {@link TestStat} to be associated with the deck
	 * @return <code>true</code> if the {@link TestStat} is inserted, <code>false</code> otherwise
	 */
	public boolean addTestStat(TestStat d)
	{
		hasChanged = true;
		return this.testStats.add(d);
	}

	/**
	 * returns the current cardIdIterator value and increments it
	 * 
	 * @return the unused ID
	 */
	private int getUnusedID()
	{
		return ++cardIdIterator;
	}

	/**
	 * Removes all {@link Card}s from the <code>Deck</code>
	 * 
	 * @see java.util.Collection#clear()
	 */
	public void clearCards()
	{
		hasChanged = true;
		this.cards.clear();
	}

	/**
	 * Clears all of the statistics associated with this <code>Deck</code>
	 * 
	 * @see java.util.Collection#clear()
	 */
	public void clearTestStats()
	{
		hasChanged = true;
		this.testStats.clear();
	}

	/**
	 * Returns a map of cards to probabilities for the ELevel Challenge quiz.
	 * The higher the number of times a deck was incorrectly answered the higher the probability in the map.
	 * 
	 * @return a <code>Map<Card, Float></code> of <code>Card</code>s and their associated probability
	 */
	public Map<Card, Float> getELevelChallengeDeck()
	{
		Map<Card, Float> elvlDeck = new HashMap<Card, Float>();
		for(Card card : cards)
		{
			elvlDeck.put(card, 0.0F);
		}
		Date currentDate = new Date(System.currentTimeMillis());
		Date previousDate = new Date(System.currentTimeMillis() - 604800);
		int totalIncorrect = 0;
		for(TestStat stat : testStats)
		{
			// if the statistic set is over 7 days old we ignore it.
			if(currentDate.compareTo(stat.getDate()) >= 0 && previousDate.compareTo(stat.getDate()) <= 0)
			{
				for(String cardID : stat.getIncorrectlyAnsweredCards())
				{
					// TODO: need a better method of identifying cards
					for(Card card : cards)
					{
						if(card.getQuestion().compareTo(cardID) == 0)
						{
							elvlDeck.put(card, elvlDeck.get(card) + 1);
							totalIncorrect++;
						}
					}
				}
			}
		}

		if(totalIncorrect == 0)
		{
			for(Card card : cards)
			{
				elvlDeck.put(card, 1.0F / (float) cards.size());
			}
		}
		else
		{
			for(Card card : cards)
			{
				elvlDeck.put(card, elvlDeck.get(card) / (float) totalIncorrect);
			}
		}

		return elvlDeck;
	}

	/**
	 * Returns <code>true</code> if this <code>Deck</code> contains all specified {@link Card}s
	 * 
	 * @param cards
	 *        A <code>Collection<code> to be checked for containment in this collection
	 * @see java.util.Collection#containsAll(Collection)
	 */
	public boolean containsAllCards(Collection<Card> cards)
	{
		return this.cards.containsAll(cards);
	}

	/**
	 * Returns <code>true</code> if this <code>Deck</code> contains the specified {@link Card}
	 * 
	 * @param c
	 *        the {@link Card} whose presence in this <code>Deck</code> is to be tested
	 * @see java.util.Collection#contains(Object)
	 */
	public boolean containsCard(Card c)
	{
		return this.cards.contains(c);
	}

	/**
	 * Returns an iterator over this {@code Deck}'s statistics
	 * 
	 * @return an iterator over the collection of {@link TestStat}s of this {@code Deck}
	 * @see java.util.Collection#iterator()
	 */
	public Iterator<TestStat> testStatisticsIterator()
	{
		return this.testStats.iterator();
	}

	public boolean equals(Deck other)
	{
		return this.hashCode() == other.hashCode();
	}

	/**
	 * Returns the {@link Collection} of {@link Card}s in this <code>Deck</code>
	 * 
	 * @return the {@link Collection} of {@link Card}s in this <code>Deck</code>
	 */
	public List<Card> getCards()
	{
		return this.cards;
	}

	/**
	 * Returns the {@link Card} at the index of this deck
	 * 
	 * @param index
	 * @return The card at the index of this deck
	 */
	public Card getCard(int index)
	{
		return cards.get(index);
	}

	/**
	 * Returns a {@link Card} representing the currently selected card
	 * 
	 * @return a {@link Card} representing the currently selected card
	 */
	public Card getCurrentCard()
	{
		return this.cards.get(this.currentCard);
	}

	/**
	 * Returns a {@link Collection} of {@link TestStat}s representing the cumulative statistics of this deck
	 * 
	 * @return a {@link Collection} of {@link TestStat}s representing the cumulative statistics of this deck
	 */
	public Collection<TestStat> getTestStatistics()
	{
		return this.testStats;
	}

	/**
	 * Returns a {@code String} representing the location of this {@code Deck} on disk
	 * 
	 * @return a {@code String} representing the location of this {@code Deck} on disk
	 */
	public String getDiskLocation()
	{
		return this.diskLocation;
	}

	/**
	 * Returns a <code>String</code> representing the name of this <code>Deck</code>
	 * 
	 * @return a <code>String</code> representing the name of this <code>Deck</code>
	 */
	public String getName()
	{
		return this.name;
	}

	public int hashCode()
	{
		return ((this.name != null) ? 97 * this.name.hashCode() : 0) + ((this.testStats != null) ? 79 * this.testStats.hashCode() : 0) + ((this.diskLocation != null) ? 61 * diskLocation.hashCode() : 0) + ((this.cards != null) ? 47 * this.cards.hashCode() : 0) + (3 * this.currentCard) + ((hasChanged) ? 1 : 0);
	}

	/**
	 * Returns the index of the specified <code>Card</code> in the specified <code>Deck</code>
	 * 
	 * @return an <code>int</code> representing the index of the specified <code>Card</code> in the specified
	 *         <code>Deck</code>, -1 if the <code>Card</code> is not contained in the specified <code>Deck</code>
	 */
	public int indexOf(Card c)
	{
		if(cards.contains(c))
			return cards.indexOf(c);
		else
			return -1;
	}

	/**
	 * Returns <code>true</code> if there are no statistics
	 * 
	 * @return <code>true</code> if there are no statistics
	 * @see java.util.Collection#isEmpty()
	 */
	public boolean isDeckStatEmpty()
	{
		return this.testStats.isEmpty();
	}

	/**
	 * Returns the number of cards in this deck
	 * 
	 * @return the number of cards in this deck
	 */
	public int numCards()
	{
		return this.cards.size();
	}

	/**
	 * Returns the number of test statistics in this deck
	 * 
	 * @return the number of test statistics in this deck
	 * @see java.util.Collection#size()
	 */
	public int numDeckStatistics()
	{
		return this.testStats.size();
	}

	/**
	 * Read a <code>Deck</code> from disk given an absolute file name
	 * 
	 * @param parent
	 *        A <code>QWidget</code> representing the parent <code>QWidget</code>
	 * @param filepath
	 *        A <code>String</code> representing the absolute path file name of the file
	 * @return a {@link Deck} created from the specified file
	 */
	public static Deck readFromDisk(QWidget parent, String filepath)
	{
		try
		{
			ObjectInputStream in = new ObjectInputStream(new FileInputStream(filepath));
			Deck d = (Deck) in.readObject();
			in.close();

			if(d != null)
			{
				// set the first card to the "current" card
				d.updateCurrentCard();
				d.setDiskLocation(filepath);
				d.setHasChanged(false);
			}
			return d;
		}
		catch(FileNotFoundException e)
		{
			e.printStackTrace();
			QMessageBox.critical(parent, "FileNotFoundException", "The specified file was not found!");
		}
		catch(IOException e)
		{
			e.printStackTrace();
			QMessageBox.critical(parent, "IOException", "An I/O error has occurred!");
		}
		catch(ClassNotFoundException e)
		{
			e.printStackTrace();
			QMessageBox.critical(parent, "ClassNotFoundException", "I don't know what a Deck is!");
		}
		return null;
	}

	/**
	 * Removes the specified {@link Card} from the <code>Deck</code>, if it is present
	 * 
	 * @param c
	 *        the {@link Card} to be removed from the <code>Deck</code>, if it is present
	 * @see java.util.Collection#remove(Object)
	 */
	public boolean removeCard(Card c)
	{
		hasChanged = true;
		return this.cards.remove(c);
	}

	/**
	 * Gets the next <code>Card</code> in the deck
	 * 
	 * @return the next card in the deck after the current card
	 */
	public Card selectNextCard()
	{
		if(this.currentCard + 1 < this.cards.size())
			this.setCurrentCard(this.currentCard + 1);
		else
			this.setCurrentCard(0);

		return this.cards.get(this.currentCard);
	}

	/**
	 * Gets the previous <code>Card</code> in the deck
	 * 
	 * @return the previous card in the deck before the current card
	 */
	public Card selectPreviousCard()
	{
		if(this.currentCard > 0)
			this.setCurrentCard(this.currentCard - 1);
		else
			this.setCurrentCard(this.cards.size() - 1);

		return this.cards.get(this.currentCard);
	}

	/**
	 * Updates the currently selected {@code Card} to the specified {@link Card}
	 * 
	 * @param c
	 *        a {@link Card} representing the currently selected {@link Card}
	 */
	public void setCurrentCard(Card c)
	{
		int i = this.cards.lastIndexOf(c);
		if(i < 0)
			this.currentCard = 0;
		else
			this.currentCard = i;
	}

	/**
	 * Updates the currently selected <code>Card</code> to the <code>Card</code> at the specified index
	 * 
	 * @param i
	 *        an <code>int</code> representing the index of the currently selected {@link Card}
	 */
	public void setCurrentCard(int i)
	{
		this.currentCard = i;
	}

	/**
	 * Updates the location of this <code>Deck</code> on disk
	 * 
	 * @param d
	 *        a <code>String</code> representing the new location of this <code>Deck</code> on disk
	 */
	public void setDiskLocation(String d)
	{
		hasChanged = true;
		this.diskLocation = d;
	}

	/**
	 * Sets the hasChanged flag to the specified <code>boolean</code>
	 * 
	 * @param b
	 *        the <code>boolean</code> to update the flag to
	 */
	public void setHasChanged(boolean b)
	{
		hasChanged = b;
	}

	/**
	 * @return The value of the hasChanged field.
	 */
	public boolean hasChanged()
	{
		return hasChanged;
	}

	/**
	 * Updates the name of this <code>Deck</code> to the specified <code>String</code>
	 * 
	 * @param n
	 *        a <code>String</code> representing the new name of this <code>Deck</code>
	 */
	public void setName(String n)
	{
		if(n == null || n.equals(""))
		{
			this.name = "Default Deck";
			hasChanged = true;
		}
		else
			this.name = n;
	}

	/**
	 * Randomizes the ordering of the cards in the card list, "shuffling" them
	 */
	public void shuffle()
	{
		if(cards.size() <= 1)
			return;

		Card current = this.cards.get(this.currentCard);

		ArrayList<Card> reordered = new ArrayList<Card>();
		Random rand = new Random();

		// Remove a uniformly distributed card from the current card list and add it to a temporary list
		while(this.cards.size() > 0)
			reordered.add(this.cards.remove(rand.nextInt(this.cards.size())));

		this.cards = reordered;
		this.currentCard = this.cards.indexOf(current);

		hasChanged = true;
	}

	/**
	 * Sets the current card to the first card of the deck
	 */
	public void updateCurrentCard()
	{
		if(this.cards != null && this.cards.size() > 0)
			this.currentCard = 0;
	}

	/**
	 * Export the contents of this <code>Deck</code> to disk
	 * 
	 * @param parent
	 *        A <code>QWidget</code> representing the parent <code>QWidget</code>
	 * @param filepath
	 *        A <code>String</code> representing the absolute path file name of the file
	 */
	public void writeToDisk(QWidget parent, String filepath)
	{
		try
		{
			if(!filepath.endsWith(".elvl"))
				filepath += ".elvl";
			ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(filepath));
			this.diskLocation = new String(filepath);
			hasChanged = false;
			out.writeObject(this);
			out.close();
		}
		catch(IOException e)
		{
			QMessageBox.critical(parent, "IOException", "An I/O error has occurred!");
			e.printStackTrace();
		}
	}

	/**
	 * A <code>String</code> representation of the entire deck (including cards).
	 */
	public String toString()
	{
		String r = "";
		r += this.name + "\n";
		r += this.diskLocation + "\n";
		r += this.hasChanged + "\n";
		r += this.currentCard + "\n";
		for(Card c : cards)
			r += c.toString();
		for(TestStat t : testStats)
			r += t.toString();
		return r;
	}
}
