/**
 * DeckStat.java
 * Group 7, Object Oriented Software Engineering
 * October 10, 2009
 */
package edu.jhu.cs.oose.elevel.dataModel;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * Contains statistics for a single test of a {@link Deck}
 * 
 * @author Jesse Yates, H. Parker Shelton, Greg Anderson
 */
public class TestStat implements Serializable
{
	private static final long serialVersionUID = 8469838483846584L;

	/** The number of questions answered in this test */
	private int numAnswered;

	/** The number of questions answered correctly in this test */
	private int numCorrect;

	private List<String> incorrectlyAnswered;

	// Date = new Date() sets it to the number of seconds since "epoch"
	/** The date and time the test began */
	private Date date;

	/** The duration of this test in seconds */
	private int duration;

	/**
	 * Initializes an empty <code>DeckStat</code> with the current date and time
	 */
	public TestStat(long currentTimeMillis)
	{
		this(0, 0, new Date(currentTimeMillis), 0);
	}

	/**
	 * Initializes a <code>DeckStat</code> with the specified parameters
	 * 
	 * @param numA
	 *        an <code>int</code> representing the number of questions answered in a particular test
	 * @param numC
	 *        an <code>int</code> representing the number of questions answered correctly in a particular test
	 * @param d
	 *        an {@link Date} representing the date and time of a particular test
	 * @param duration
	 *        an <code>int</code> representing the duration in seconds of a particular test
	 */
	public TestStat(int numA, int numC, Date d, int duration)
	{
		this.numAnswered = numA;
		this.numCorrect = numC;
		this.date = d;
		this.duration = duration;

		incorrectlyAnswered = new ArrayList<String>();
	}

	public boolean equals(TestStat other)
	{
		return this.hashCode() == other.hashCode();
	}

	/**
	 * Returns a <code>String</code> representing the date and time of this test
	 * 
	 * @return a <code>String</code> representing the date and time of this test
	 */
	public Date getDate()
	{
		return date;
	}

	/**
	 * Returns the test's duration in seconds
	 * 
	 * @return an <code>int</code> representing the duration of the test
	 */
	public int getDuration()
	{
		return this.duration;
	}

	/**
	 * Returns a <code>double</code> representing the percent accuracy on this test
	 * 
	 * @return a <code>double</code> representing the percent accuracy on this test
	 */

	public double getPercentCorrect()
	{
		return (numAnswered != 0) ? ((double) numCorrect) / numAnswered : 0;
	}

	/**
	 * Returns the hashcode of this <code>TestStat</code>.
	 */
	public int hashCode()
	{
		return (97 * this.numAnswered) + (79 * numCorrect) + ((this.incorrectlyAnswered != null) ? 61 * this.incorrectlyAnswered.hashCode() : 0) + ((this.date != null) ? 47 * this.date.hashCode() : 0) + (3 * this.duration);
	}

	/**
	 * Sets the duration of the test to the specified <code>int</code>
	 * 
	 * @param t
	 *        an <code>int</code> representing the length of time the test took
	 */
	public void setDuration(int t)
	{
		this.duration = t;
	}

	/**
	 * Sets the date of the test to the specified {@link Date}
	 * 
	 * @param d
	 *        a {@link Date} representing the date and time of the test
	 */
	public void setDate(Date d)
	{
		this.date = d;
	}

	/**
	 * Sets the date of the test to the specified <code>int</code>
	 * 
	 * @param d
	 *        a <code>int</code> representing the date and time of the test in seconds since Jan 1, 1970
	 */
	public void setDate(int d)
	{
		this.date = new Date(d);
	}

	/**
	 * Increments the overall number of questions answered and the number correctly answered if
	 * the answer was correct
	 * 
	 * @param correct
	 *        <code>true</code> if the question was answered correctly during the test, <code>false</code> otherwise
	 */
	public void incrementPercent(boolean correct)
	{
		this.numAnswered++;
		if(correct)
			this.numCorrect++;
	}

	/**
	 * Adds an incorrectly answered card to a list for use with the E-level Challenge
	 * 
	 * @param card
	 *        the card to add
	 */
	public void addIncorrectlyAnsweredCard(String card)
	{
		if(!incorrectlyAnswered.contains(card))
			incorrectlyAnswered.add(card);
	}

	/**
	 * returns the list of incorrectly answered cards
	 * 
	 * @return a <code>List</code> of incorrectly answered cards
	 */
	public List<String> getIncorrectlyAnsweredCards()
	{
		// TODO: incorrectly answered card iterator?
		return incorrectlyAnswered;
	}

	/**
	 * Format a <code>DeckStat</code> to display at the end of a test
	 * 
	 * @return a <code>String</code> representation of the <code>DeckStat</code>
	 */
	public String toString()
	{
		String r = "";
		r += numCorrect + "/" + numAnswered + " answered correctly\n";
		r += "Duration: " + duration + " seconds\n";
		return r;
	}
}
