/**
 * PreferenceWindow.java
 * Group 7, Object Oriented Software Engineering
 * October 19, 2009
 */
package edu.jhu.cs.oose.elevel.preferences;

import com.trolltech.qt.core.Qt;
import com.trolltech.qt.gui.QCloseEvent;
import com.trolltech.qt.gui.QDialog;
import com.trolltech.qt.gui.QGridLayout;
import com.trolltech.qt.gui.QHBoxLayout;
import com.trolltech.qt.gui.QKeyEvent;
import com.trolltech.qt.gui.QLabel;
import com.trolltech.qt.gui.QListWidget;
import com.trolltech.qt.gui.QPushButton;
import com.trolltech.qt.gui.QStackedWidget;
import com.trolltech.qt.gui.QVBoxLayout;
import com.trolltech.qt.gui.QWidget;
import edu.jhu.cs.oose.elevel.utilities.PreferenceCheckBox;

/**
 * Creates a Window in which the current program preference can be
 * changed and viewed
 * 
 * @author Jesse Yates, H. Parker Shelton
 */
public class PreferenceWindow extends QDialog
{
	/** QWidgets for the different preference panes */
	QWidget program;
	QWidget testing;

	/** The QLayout for the window */
	QStackedWidget stackedWidget;

	/** The QListWidget for the window */
	QListWidget listWidget;

	/** A QLabel that stores the text of the default export location */
	QLabel exportLabel;

	/**
	 * Initializes a <code>PreferenceWindow</code> with the given parent
	 * 
	 * @param parent
	 *        A <code>QWidget</code> representing the parent window
	 */
	public PreferenceWindow(QWidget parent)
	{
		super(parent);

		listWidget = new QListWidget();
		listWidget.addItem(tr("Program Settings"));
		listWidget.addItem(tr("Test Settings"));
		listWidget.setFixedWidth(120);

		// QPushButton applyButton = new QPushButton(tr("Apply"));
		QPushButton closeButton = new QPushButton(tr("Close"));
		closeButton.setDefault(true);
		QPushButton revertButton = new QPushButton(tr("Restore to Defaults"));

		QHBoxLayout buttonsLayout = new QHBoxLayout();
		buttonsLayout.addWidget(revertButton);
		buttonsLayout.addStretch(1);
		// buttonsLayout.addWidget(applyButton);
		buttonsLayout.addWidget(closeButton);

		revertButton.pressed.connect(this, "restoreDefaults()");
		closeButton.pressed.connect(this, "close()");
		// closeButton.pressed.connect(this, "restorePrevious()");
		// applyButton.pressed.connect(this, "savePreferences()");

		this.stackedWidget = new QStackedWidget();
		listWidget.currentRowChanged.connect(this.stackedWidget, "setCurrentIndex(int)");
		listWidget.setCurrentRow(0);

		this.program = createProgramPreferences();
		this.testing = createTestingPreferences();
		this.stackedWidget.addWidget(this.program);
		this.stackedWidget.addWidget(this.testing);

		QHBoxLayout horizontalLayout = new QHBoxLayout();
		horizontalLayout.addWidget(listWidget);
		horizontalLayout.addWidget(stackedWidget);

		QVBoxLayout windowLayout = new QVBoxLayout();
		windowLayout.addLayout(horizontalLayout);
		windowLayout.addLayout(buttonsLayout);
		this.setLayout(windowLayout);

		setWindowTitle(tr("Preferences"));
	}

	/**
	 * Displays a dialog to change the default export directory
	 */
	public void changeExportLocation()
	{
		Preferences.getInstance().changeExportLocation(this, "Please select a new directory for E-Level decks:");
		exportLabel.setText(truncatePath(Preferences.getInstance().getValue("DEFAULT_EXPORT_LOCATION")));
	}

	/**
	 * Truncates the given string, returning the root directory and the last two directories
	 * 
	 * @param s
	 *        a <code>String</code> representing the path of a directory
	 * @return
	 *         a <code>String</code> with the first and last two directories and ellipses in the middle
	 */
	private String truncatePath(String path)
	{
		String s = path;
		if(s.length() > 15)
		{
			if(s.endsWith("/"))
				s = s.substring(0, s.length() - 1);
			String end = s.substring(s.substring(0, s.lastIndexOf("/")).lastIndexOf("/"));
			String start = s.substring(0, s.substring(1).indexOf("/") + 2);
			s = start + "..." + end;
		}
		return s;
	}

	/**
	 * Writes out the preferences when the window is closed
	 */
	public void closeEvent(QCloseEvent event)
	{
		Preferences.getInstance().writeSettings();
	}

	public void keyPressEvent(QKeyEvent event)
	{
		if(event.key() == Qt.Key.Key_Escape.value())
			this.close();
	}

	/**
	 * Initializes a widget containing the preferences for the program overall
	 */
	private QWidget createProgramPreferences()
	{
		QGridLayout layout = new QGridLayout();
		layout.addWidget(new QLabel("<b>Program Settings</b>"), 0, 0, 1, 2, new Qt.Alignment(Qt.AlignmentFlag.AlignCenter, Qt.AlignmentFlag.AlignTop));

		// Add checkbox for the splash screen
		PreferenceCheckBox splash = new PreferenceCheckBox(tr("Enable splash screen"), "SHOW_SPLASH");
		splash.setChecked(Preferences.getInstance().getValue("SHOW_SPLASH").equals("true"));
		splash.stateChange.connect(this, "updatePreference(String, boolean)");
		layout.addWidget(splash, 1, 0);

		// Add checkbox for open previously open decks
		PreferenceCheckBox openLast = new PreferenceCheckBox(tr("Open previously open decks at startup"), "OPEN_PREVIOUS");
		openLast.setChecked(Preferences.getInstance().getValue("OPEN_PREVIOUS").equals("true"));
		openLast.stateChange.connect(this, "updatePreference(String, boolean)");
		layout.addWidget(openLast, 2, 0);

		// Add text for default export location
		exportLabel = new QLabel(truncatePath(Preferences.getInstance().getValue("DEFAULT_EXPORT_LOCATION")));
		layout.addWidget(exportLabel, 3, 0, Qt.AlignmentFlag.AlignLeft);

		// Add button for default export location
		QPushButton b = new QPushButton(tr("Export Location"), this);
		b.clicked.connect(this, "changeExportLocation()");
		layout.addWidget(b, 3, 1, Qt.AlignmentFlag.AlignCenter);

		QWidget programPage = new QWidget();
		programPage.setLayout(layout);
		return programPage;
	}

	/**
	 * Initializes a widget containing the preferences for testing
	 */
	private QWidget createTestingPreferences()
	{
		QGridLayout layout = new QGridLayout();
		layout.addWidget(new QLabel("<b>Test Settings</b>"), 0, 0, 1, 2, new Qt.Alignment(Qt.AlignmentFlag.AlignCenter, Qt.AlignmentFlag.AlignTop));

		// Add checkbox for show timer
		PreferenceCheckBox showTimer = new PreferenceCheckBox(tr("Show the timer"), "SHOW_TIMER");
		showTimer.setChecked(Preferences.getInstance().getValue("SHOW_TIMER").equals("true"));
		showTimer.stateChange.connect(this, "updatePreference(String, boolean)");
		layout.addWidget(showTimer, 1, 0, Qt.AlignmentFlag.AlignTop);

		// Add checkbox for show score
		PreferenceCheckBox showScore = new PreferenceCheckBox(tr("Show the current percent correct"), "SHOW_SCORE");
		showScore.setChecked(Preferences.getInstance().getValue("SHOW_SCORE").equals("true"));
		showScore.stateChange.connect(this, "updatePreference(String, boolean)");
		layout.addWidget(showScore, 2, 0, Qt.AlignmentFlag.AlignTop);

		// Add checkbox for allow hints
		PreferenceCheckBox allowHints = new PreferenceCheckBox(tr("Allow hints"), "ALLOW_HINTS");
		allowHints.setChecked(Preferences.getInstance().getValue("ALLOW_HINTS").equals("true"));
		allowHints.stateChange.connect(this, "updatePreference(String, boolean)");
		layout.addWidget(allowHints, 3, 0, Qt.AlignmentFlag.AlignTop);

		// Add checkbox for always show preferences
		PreferenceCheckBox alwaysShow = new PreferenceCheckBox(tr("Always show preferences on testing"), "SHOW_TEST_PREFERENCES");
		alwaysShow.setChecked(Preferences.getInstance().getValue("SHOW_TEST_PREFERENCES").equals("true"));
		alwaysShow.stateChange.connect(this, "updatePreference(String, boolean)");
		layout.addWidget(alwaysShow, 4, 0, Qt.AlignmentFlag.AlignTop);

		QWidget testingPage = new QWidget();
		testingPage.setLayout(layout);
		return testingPage;
	}

	/**
	 * Restores the preferences to a set of default values
	 */
	public void restoreDefaults()
	{
		if(this.listWidget.currentRow() == 0)
		{
			Preferences.getInstance().setValue("SHOW_SPLASH", "true");
			Preferences.getInstance().setValue("OPEN_PREVIOUS", "false");
			Preferences.getInstance().setDefaultLocation();
			stackedWidget.removeWidget(program);
			program = createProgramPreferences();
			stackedWidget.insertWidget(0, program);
			stackedWidget.setCurrentIndex(0);
		}
		else if(this.listWidget.currentRow() == 1)
		{
			Preferences.getInstance().setValue("SHOW_TIMER", "true");
			Preferences.getInstance().setValue("SHOW_SCORE", "true");
			Preferences.getInstance().setValue("ALLOW_HINTS", "true");
			Preferences.getInstance().setValue("SHOW_TEST_PREFERENCES", "true");
			stackedWidget.removeWidget(testing);
			testing = createTestingPreferences();
			stackedWidget.insertWidget(1, testing);
			stackedWidget.setCurrentIndex(1);
		}
	}

	/**
	 * Restores the preferences to the previous values
	 */
	public void restorePrevious()
	{
		this.close();
	}

	/**
	 * Set the preference represented by the specified key to the specified value, to be written on close.
	 * Used by the {@link PreferenceCheckBox}
	 * 
	 * @param s
	 *        A <code>String</code> representing the name of a preference
	 * @param b
	 *        A <code>boolean</code> representing the value of the preference
	 */
	public void updatePreference(String s, boolean b)
	{
		Preferences.getInstance().setValue(s, String.valueOf(b));
	}
}
