/**
 * Preferences.java
 * Group 7, Object Oriented Software Engineering
 * October 14, 2009
 */
package edu.jhu.cs.oose.elevel.preferences;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import com.trolltech.qt.QSysInfo;
import com.trolltech.qt.core.QDir;
import com.trolltech.qt.core.QSettings;
import com.trolltech.qt.gui.QFileDialog;
import com.trolltech.qt.gui.QWidget;
import edu.jhu.cs.oose.elevel.dataModel.Deck;

/**
 * Stores preferences for the E-Level application, implementing {@link QSettings}.
 * 
 * @author H. Parker Shelton, Jesse Yates
 */
public class Preferences
{
	/** Map of all the preferences and their values */
	Map<String, String> preferences;

	/** Company name used to identify the program */
	private static final String COMPANY_NAME = "OOSE";

	/** Product name used to identify the program */
	private static final String PRODUCT_NAME = "E_LEVEL";

	/** Singleton instance of the class */
	private static final Preferences INSTANCE = new Preferences();

	/**
	 * Get the singleton instance of this class
	 * 
	 * @return an instance of this class
	 */
	public static Preferences getInstance()
	{
		return INSTANCE;
	}

	/**
	 * Initializes a new Preferences. If an external file already exists, it is populated
	 * with the existing data. Otherwise, it is set to default preferences.
	 */
	private Preferences()
	{
		this.preferences = new HashMap<String, String>();
		readExternalSettings();

		if(this.preferences.size() == 0)
			setDefault(true);
		else
			setDefault(false);
	}

	/**
	 * Prompts the user for input if there is not a default directory
	 */
	public void changeExportLocation(QWidget parent, String text)
	{
		String initialPath = hasExportLocation() ? (String) Preferences.getInstance().getValue("DEFAULT_EXPORT_LOCATION") : QDir.homePath().toString();
		String path = QFileDialog.getExistingDirectory(parent, parent.tr(text), initialPath);
		if(!(path == null || path.equals("")))
		{
			// Assign a new path
			if(!path.endsWith("/"))
				path = path + "/";
			Preferences.getInstance().setValue("DEFAULT_EXPORT_LOCATION", path);
		}
		else
		{
			// User canceled
			if(hasExportLocation())
			{
				// Revert to previous location
				Preferences.getInstance().setValue("DEFAULT_EXPORT_LOCATION", initialPath);
			}
			else
				setDefaultLocation();

			File f = new File(Preferences.getInstance().getValue("DEFAULT_EXPORT_LOCATION"));
			if(!f.exists())
				f.mkdir();
		}
		writeSettings();
	}

	/**
	 * Returns a <code>Map</code> containing all the testing preferences
	 * 
	 * @return a <code>Map</code> containing all the testing preferences
	 */
	public Map<String, String> getTestingPreferences()
	{
		Map<String, String> m = new HashMap<String, String>();
		m.put("SHOW_TIMER", Preferences.getInstance().getValue("SHOW_TIMER"));
		m.put("ALLOW_HINTS", Preferences.getInstance().getValue("ALLOW_HINTS"));
		m.put("SHOW_SCORE", Preferences.getInstance().getValue("SHOW_SCORE"));
		m.put("SHOW_TEST_PREFERENCES", Preferences.getInstance().getValue("SHOW_TEST_PREFERENCES"));
		return m;
	}

	/**
	 * Get the preference with the specified name
	 * 
	 * @param key
	 *        A <code>String</code> representing the name of the preference
	 * @return
	 *         An <code>Object</code> representing the value of the specified preference
	 */
	public String getValue(String key)
	{
		return this.preferences.get(key);
	}

	/**
	 * Returns whether or not the program has a default export location that exists on disk.
	 * 
	 * @return <code>true</code> if the program has a default export location and the directory exists on disk,
	 *         <code>false</code> otherwise.
	 */
	public boolean hasExportLocation()
	{
		if(Preferences.getInstance().getValue("DEFAULT_EXPORT_LOCATION") == null || Preferences.getInstance().getValue("DEFAULT_EXPORT_LOCATION").equals(""))
			return false;

		// TODO Decide the correct behavior
		String loc = (String) Preferences.getInstance().getValue("DEFAULT_EXPORT_LOCATION");
		loc = loc.substring(0, loc.lastIndexOf('/')).substring(0, loc.lastIndexOf('/'));
		File f = new File(loc);
		if(!f.exists())
			// f.mkdir(); // Recreate a directory that does not currently exist on disk
			return false; // Have the user choose a new directory
		return true;
	}

	/**
	 * Read the settings in from the configuration file on disk
	 * 
	 * @see QSettings
	 */
	private void readExternalSettings()
	{
		QSettings settings = new QSettings(COMPANY_NAME, PRODUCT_NAME);

		settings.beginGroup("Preferences");

		for(String key : settings.allKeys())
			this.preferences.put(key, (String) settings.value(key, ""));

		settings.endGroup();
	}

	/**
	 * Returns a list of file locations for the previously open decks
	 */
	public List<String> getPreviouslyOpenDecks()
	{
		List<String> l = new ArrayList<String>();
		QSettings settings = new QSettings(COMPANY_NAME, PRODUCT_NAME);

		settings.beginGroup("Open Decks");
		for(String key : settings.allKeys())
			l.add((String) settings.value(key));
		settings.endGroup();

		return l;
	}

	/**
	 * Set the default settings on the first run
	 */
	public void setDefault(boolean overwrite)
	{
		if(overwrite)
		{
			setValue("SHOW_SPLASH", "true");
			setValue("DEFAULT_EXPORT_LOCATION", "");
			setValue("OPEN_PREVIOUS", "false");
			setValue("SHOW_TIMER", "true");
			setValue("ALLOW_HINTS", "true");
			setValue("SHOW_SCORE", "true");
			setValue("SHOW_TEST_PREFERENCES", "true");
		}
		else
		{
			if(!preferences.containsKey("SHOW_SPLASH"))
				setValue("SHOW_SPLASH", "true");
			if(!preferences.containsKey("DEFAULT_EXPORT_LOCATION"))
				setDefaultLocation();
			if(!preferences.containsKey("OPEN_PREVIOUS"))
				setValue("OPEN_PREVIOUS", "false");
			if(!preferences.containsKey("SHOW_TIMER"))
				setValue("SHOW_TIMER", "true");
			if(!preferences.containsKey("SHOW_SCORE"))
				setValue("SHOW_SCORE", "true");
			if(!preferences.containsKey("ALLOW_HINTS"))
				setValue("ALLOW_HINTS", "true");
			if(!preferences.containsKey("SHOW_TEST_PREFERENCES"))
				setValue("SHOW_TEST_PREFERENCES", "true");
		}
	}

	/**
	 * Set the default export location to a platform-dependent location on disk
	 */
	public void setDefaultLocation()
	{
		// Assign an arbitrary default location
		if(QSysInfo.windowsVersion() != 0) // Windows
			setValue("DEFAULT_EXPORT_LOCATION", QDir.homePath().toString() + "/My Documents/E-Level/");
		else if(QSysInfo.macVersion() != 0) // Mac OS X
			setValue("DEFAULT_EXPORT_LOCATION", QDir.homePath().toString() + "/Documents/E-Level/");
		else
			setValue("DEFAULT_EXPORT_LOCATION", QDir.homePath().toString() + "/Documents/E-Level/");
	}

	/**
	 * Set the preference represented by the specified key to the specified value, to be written on close
	 * 
	 * @param key
	 *        A <code>String</code> representing the name of a preference
	 * @param value
	 *        A <code>Object</code> representing the value of the preference
	 */
	public void setValue(String key, String value)
	{
		this.preferences.put(key, value);
	}

	/**
	 * Write all the current settings out to a file on disk
	 */
	public void writeSettings()
	{
		QSettings settings = new QSettings(COMPANY_NAME, PRODUCT_NAME);
		settings.remove("Open Decks");

		settings.beginGroup("Preferences");
		for(Entry<String, String> e : this.preferences.entrySet())
			settings.setValue(e.getKey(), e.getValue());
		settings.endGroup();
	}

	/**
	 * Write all the current settings out to a file on disk
	 */
	public void writeSettings(List<Deck> openDecks)
	{
		QSettings settings = new QSettings(COMPANY_NAME, PRODUCT_NAME);

		settings.remove("Open Decks");
		settings.beginGroup("Open Decks");
		for(Deck d : openDecks)
			settings.setValue(d.getName(), d.getDiskLocation());
		settings.endGroup();
	}
}
