/**
 * TestWindow.java
 * Group 7, Object Oriented Software Engineering
 * October 10, 2009
 */
package edu.jhu.cs.oose.elevel.testTaking;

import java.util.HashMap;
import java.util.Map;
import edu.jhu.cs.oose.elevel.dataModel.Card;
import edu.jhu.cs.oose.elevel.dataModel.Deck;
import edu.jhu.cs.oose.elevel.dataModel.TestStat;
import edu.jhu.cs.oose.elevel.preferences.Preferences;

/**
 * The underlying data model to be displayed in the associated {@link TestWindowGUI}
 * 
 * @author Jesse Yates, H. Parker Shelton
 */
public class TestWindow
{
	/** The <code>Deck</code> currently being tested */
	protected Deck deck;

	/** The statistics for the current test */
	private TestStat testStat;

	/** The number of cards in the deck */
	private int numCards;

	/** The number of cards tested */
	private int cardsTested;

	/** The start time of the current test */
	protected long startTime;

	/** The set of preferences for the given test */
	private Map<String, String> preferences;

	/**
	 * Initializes a new <code>TestWindow</code> with the specified <code>Deck</code>
	 * 
	 * @param currentDeck
	 *        a <code>Deck</code> representing the currently selected <code>Deck</code> on which to test
	 */
	public TestWindow(Deck currentDeck)
	{
		preferences = new HashMap<String, String>();
		preferences.putAll(Preferences.getInstance().getTestingPreferences());
		preferences.put("TEST_FRONT_TO_BACK", "true");

		this.deck = currentDeck;
		this.numCards = this.deck.numCards();
		initializeTest();
	}

	/**
	 * Gets the <code>String</code> value of the preference denoted by a <code>String</code>
	 * 
	 * @param s
	 *        The preference of interest.
	 * @return The value of the preference of interest.
	 */
	protected String getPreference(String s)
	{
		return this.preferences.get(s);
	}

	/**
	 * Sets the value of the preference denoted by a <code>String</code>
	 * 
	 * @param k
	 *        The preference to set.
	 * @param v
	 *        The value of the preference.
	 */
	protected void updatePreference(String k, boolean v)
	{
		this.preferences.put(k, String.valueOf(v));
	}

	/**
	 * Sets the currently selected <code>Deck</code> to the specified {@link Deck}
	 * 
	 * @param d
	 *        a <code>Deck</code> to assign to the currently selected deck
	 */
	public void setDeck(Deck d)
	{
		this.deck = d;
	}

	/**
	 * Check the given answer with the answer stored in the <code>Deck</code>
	 * 
	 * @param s
	 *        the <code>String</code> to check
	 */
	public void checkAnswer(String s)
	{
		checkAnswer(this.deck.getCurrentCard().checkAnswer(s));
	}

	/**
	 * Determine if the answer given to the previous {@link Card} was correct.
	 * Used for <code>Card</code>s with no computer-checkable answer.
	 * 
	 * @param answer
	 *        <code>true</code> if the previous answer was correct, <code>false</code> otherwise.
	 */
	private void checkAnswer(boolean answer)
	{
		testStat.incrementPercent(answer);
		if(!answer)
			testStat.addIncorrectlyAnsweredCard(this.deck.getCurrentCard().question);
	}

	/**
	 * Increment the number of correct answers
	 */
	public void rightAnswer()
	{
		checkAnswer(true);
	}

	/**
	 * Increment the number of incorrect answers and adds the current card to the list of incorrectly answered cards.
	 * Note that having answered a card incorrectly once is a necessary and sufficient condition for adding to this
	 * list.
	 */
	public void wrongAnswer()
	{
		checkAnswer(false);
	}

	/**
	 * Completes the test and permanently associates the current statistics with the current <code>Deck</code>
	 */
	public TestStat finishTest()
	{
		this.deck.selectNextCard();

		// XXX: HPS - probably not safe to make the cast, but that's a reeaally long time.
		this.testStat.setDuration(testTime());
		deck.addTestStat(this.testStat);
		return this.testStat;
		// Other closing things
	}
	
	/**
	 * Get the current duration of the test
	 * @return
	 */
	public int testTime()
	{
		return (int) Math.round(((System.currentTimeMillis() - startTime) / 1000));
	}

	/**
	 * Initializes default parameters for a new test
	 */
	protected void initializeTest()
	{
		this.cardsTested = 0;
		this.testStat = new TestStat(System.currentTimeMillis());
		this.startTime = System.currentTimeMillis();
	}

	/**
	 * Selects the next <code>Card</code>
	 * 
	 * @return <code>true</code> if there is a next <code>Card</code>, <code>false</code> if the test is over
	 */
	public boolean nextCard()
	{
		cardsTested++;
		if(cardsTested >= numCards)
			return false;
		this.deck.selectNextCard();
		return true;
	}

	/**
	 * Get the current <code>Card</code> from the {@link Deck}
	 * 
	 * @return the current <code>Card</code>
	 */
	public Card currentCard()
	{
		return this.deck.getCurrentCard();
	}
}
