package edu.jhu.cs.oose.elevel.utilities;

import java.util.ArrayList;
import java.util.List;
import com.trolltech.qt.core.QObject;
import edu.jhu.cs.oose.elevel.CardAreaGUI;
import edu.jhu.cs.oose.elevel.dataModel.Card;
import edu.jhu.cs.oose.elevel.dataModel.Deck;

/**
 * A Singleton class that keeps track of the current deck
 * and the list of decks. It also emits signals to the
 * tree and the card area when updates are required. 
 * 
 * @author Pablo Lee
 *
 */
public class ViewState extends QObject
{
	/** The single instance of this class */
	public static final ViewState INSTANCE = new ViewState();

	/** The list of decks */
	private List<Deck> decks;
	
	/** The current deck */
	private Deck currentDeck;

	/** Signals to be emitted */
	public final Signal2<Deck, Integer> deckAdded;
	public final Signal2<Deck, Integer> deckRemoved;
	public final Signal1<Deck> deckChanged;
	public final Signal2<Card, Integer> cardAdded;
	public final Signal1<Card> cardRemoved;
	public final Signal1<Card> cardSelectedInTree;
	public final Signal2<Card, Deck> cardChangedInCardArea;
	public final Signal0 treeSizeChanged;

	/**
	 * A private constructor to initialize the
	 * class. Called only once by itself
	 */
	private ViewState()
	{
		decks = new ArrayList<Deck>();
		currentDeck = null;

		deckAdded = new Signal2<Deck, Integer>();
		deckRemoved = new Signal2<Deck, Integer>();
		deckChanged = new Signal1<Deck>();
		cardAdded = new Signal2<Card, Integer>();
		cardRemoved = new Signal1<Card>();
		cardSelectedInTree = new Signal1<Card>();
		cardChangedInCardArea = new Signal2<Card, Deck>();
		treeSizeChanged = new Signal0();

	}

	/**
	 * adds {@link Deck} to the set of decks at the specified index.
	 * Doesn't add a deck without a name
	 * 
	 * @param deck the new <code>deck</code> to be added
	 * @param index the position at which to add the deck
	 */
	public void addDeckAtIndex(Deck deck, int index)
	{
		// TODO: Make sure you only have 1 deck of a type
		if(deck.getName().equals(""))
		{
			return;
		}

		currentDeck = deck;
		decks.add(index, deck);
		deckAdded.emit(deck, index);

		if(deck.numCards() > 0)
		{
			deck.setCurrentCard(deck.getCard(0));
			cardSelectedInTree.emit(deck.getCurrentCard());
		}

		treeSizeChanged.emit();
	}

	/**
	 * adds {@link Deck} to the set of decks at the end of the list.
	 * Doesn't add a deck without a name.
	 * 
	 * @param deck the new <code>deck</code> to be added
	 */
	public void addDeck(Deck deck)
	{
		addDeckAtIndex(deck, decks.size());
	}

	/**
	 * Removes {@link Deck} from the deck list
	 * 
	 * @param deck The deck to remove
	 */
	public void removeDeck(Deck deck)
	{
		int deckIndex = decks.indexOf(deck);
		deckRemoved.emit(deck, deckIndex);
		decks.remove(deck);

		if(decks.size() <= 0)
		{
			currentDeck = null;
			treeSizeChanged.emit();
		}

		if(--deckIndex < 0)
		{
			deckIndex = decks.size();
		}
		else if(decks.size() > 0)
		{
			currentDeck = decks.get(deckIndex);
		}
		
		cardSelectedInTree.emit(getCurrentCard());
	}

	/**
	 * returns the list of all decks
	 * 
	 * @return a list of decks
	 */
	public List<Deck> getDecks()
	{
		return decks;
	}

	/**
	 * Returns the number of currently open decks
	 * 
	 * @return an <code>int</code> representing the number of currently open decks
	 */
	public int getNumDecks()
	{
		return decks.size();
	}

	/**
	 * sets the current <code>deck</code> to the new deck passed in
	 * 
	 * @param deck
	 *        the new <code>deck</code>
	 */
	public void setCurrentDeck(Deck deck)
	{
		this.currentDeck = deck;
	}

	/**
	 * Shuffles the deck list
	 */
	public void shuffleCurrentDeck()
	{
		int currentCardIndex = currentDeck.getCards().indexOf(getCurrentCard());
		this.currentDeck.shuffle();
		
		currentDeck.setCurrentCard(currentCardIndex);
		
		deckChanged.emit(currentDeck);
		cardSelectedInTree.emit(getCurrentCard());
		cardChangedInCardArea.emit(getCurrentCard(), currentDeck);
	}

	/**
	 * return the current <code>deck</code> the user is on
	 * 
	 * @return the current <code>deck</code>
	 */
	public Deck getCurrentDeck()
	{
		return this.currentDeck;
	}

	/**
	 * Adds a card to the current deck at the specified
	 * index.
	 * 
	 * @param c The card to be added
	 * @param index The index at which to add the card
	 * 				into the current deck 
	 */
	public void addCardAtIndex(Card c, int index)
	{
		this.currentDeck.addCardAtIndex(index, c);
		this.currentDeck.setCurrentCard(c);
		cardAdded.emit(c, index);
	}

	/**
	 * Adds a card to the end of the current deck
	 * 
	 * @param c The card to add
	 */
	public void addCard(Card c)
	{
		addCardAtIndex(c, currentDeck.numCards());
		treeSizeChanged.emit();
	}

	/**
	 * Sets the current card in the deck, and also
	 * sets that deck as the current deck.
	 * 
	 * @param c The card to set as the current card
	 * @param d The deck the card belongs to
	 */
	public void setCurrentCard(Card c, Deck d)
	{
		setCurrentDeck(d);
		d.setCurrentCard(c);

		cardSelectedInTree.emit(c);
	}

	/**
	 * Helper method to get the current card of
	 * the current deck
	 * 
	 * @return the current card
	 */
	public Card getCurrentCard()
	{
		if(currentDeck == null)
		{
			return null;
		}
		if(currentDeck.numCards() == 0)
		{
			return null;
		}
		return this.currentDeck.getCurrentCard();
	}

	/**
	 * Removes the currently selected card
	 * from the deck.
	 */
	public void removeCurrentCard()
	{
		if(getCurrentCard() != null)
		{
			int currentCardIndex;
			Card newCurrentCard;

			if(currentDeck.numCards() > 1)
			{
				Card removingCard = getCurrentCard();
				currentCardIndex = currentDeck.getCards().indexOf(removingCard);

				if(currentCardIndex == 0)
				{
					newCurrentCard = currentDeck.getCard(1);
				}
				else
				{
					newCurrentCard = currentDeck.getCard(currentCardIndex - 1);
				}

				cardRemoved.emit(removingCard);
				currentDeck.removeCard(removingCard);
				setCurrentCard(newCurrentCard, getCurrentDeck());
				cardChangedInCardArea.emit(getCurrentCard(), getCurrentDeck());
			}
			else if(currentDeck.numCards() == 1)
			{
				Card removingCard = getCurrentCard();

				cardRemoved.emit(removingCard);
				currentDeck.removeCard(removingCard);
				cardChangedInCardArea.emit(null, getCurrentDeck());
				cardSelectedInTree.emit(null);
				treeSizeChanged.emit();
			}
		}
	}

	/**
	 * Duplicates the currently selected card
	 */
	public void duplicateCard()
	{
		Card currentCard = getCurrentCard();
		Card duplicateCard = new Card(currentCard.getQuestion(), currentCard.getAnswer(), currentCard.getHint(), currentCard.getAdditionalInfo());

		addCard(duplicateCard);
	}

	/**
	 * Causes the current card to be refreshed. Called when
	 * the current card is changed.
	 */
	public void refreshCurrentCard()
	{
		cardSelectedInTree.emit(getCurrentCard());
		cardChangedInCardArea.emit(getCurrentCard(), getCurrentDeck());
	}

	/**
	 * Sets the previous <code>Card</code> in the <code>Deck</code> to the current card and displays it. Wraps to the
	 * last card in the deck.
	 */
	@SuppressWarnings("unused")
	private void previousCard()
	{
		if(currentDeck != null)
		{
			currentDeck.selectPreviousCard();
			cardChangedInCardArea.emit(getCurrentCard(), getCurrentDeck());
		}
	}

	/**
	 * Sets the next <code>Card</code> in the <code>Deck</code> to the current card and displays it. Wraps to the first
	 * card in the deck.
	 */
	@SuppressWarnings("unused")
	private void nextCard()
	{
		if(currentDeck != null)
		{
			currentDeck.selectNextCard();
			cardChangedInCardArea.emit(getCurrentCard(), getCurrentDeck());
		}
	}
}
